/*
 * File         : BV_AudioApp.C
 * Author       : Tomasz Stachowiak
 * Created      : Fri Jul 24 18:15:10 1998
 * Revision     : $Revision$
 * Updated      : $Date$
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

#include "adpcm.h"
#include "gsm.h"

#include "BV_Application.h"
#include "BV_AudioPlay.h"
#include "BV_AudioApp.h"
#include "BV_AudioCapThread.h"
#include "BV_Session_Control.h"
#include "BV_Member.h"

BV_AudioApp::BV_AudioApp() : BV_Application(12567)
{
  int i;

  session_ = BV_Session_Control::instance();
  main_form_ = session_->getMainForm();

  BV_AudioPlay::init();
  BV_AudioCapture::init();

  for(i = 0; i < 100; i++)
    ports[i] = NULL;

  gsm_handle_ = gsm_create();

  state_ = START;
  compression_ = PCM;
  mic_cutoff = 0;

  mic_port_ = new BV_AudioCapture();
  mic_port_->open();

  cap_thread_ = new BV_AudioCapThread(this);

  cap_thread_->start();
}

BV_AudioApp::~BV_AudioApp()
{
  int i;

  cap_thread_->kill();
  //cap_thread_->stop();
  sleep(1);
  delete cap_thread_;

  mic_port_->close();
  delete mic_port_;

  for(i = 0; i < 100; i++)
    if(ports[i] != NULL)
      {
	ports[i]->close();
	delete ports[i];
	ports[i] = NULL;
      }

  gsm_destroy(gsm_handle_);

  session_->destroy();
}

int BV_AudioApp::process_message(RTPheader *header, char *body, int len)
{
  BV_Member *mbr;

  if(header->type != 2)
    {
      // wrong protocol version
      return -1;
    }

  // session_->lock();
  mbr = session_->get_member(header->ssrc);
  if(mbr == NULL)
    {
      //session_->unlock();
      return -2;   // sender not in session
    }

  int sender = mbr->get_internal_id();
  mbr->update_audio_stat(header, len);
  //session_->update_audio_stat(mbr);
  //session_->unlock();

  if(ports[sender] == NULL)
    {
      ports[sender] = new BV_AudioPlay();
      ports[sender]->open();
    }

  switch (header->pt)
    {
    case ADPCM:
      playADPCM(sender, body, len);
      break;
      
    case GSM:
      playGSM(sender, body, len);
      break;
       
    case PCM:
      //session_->lock();
      //((MainForm *)main_form_)->setSpeakerLevel(get_level(body, len/2));
      //session_->unlock();
      ports[sender]->play(body, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
      break;
      
    case STOP:
      ((MainForm *)main_form_)->setSpeakerLevel(0);
      if(ports[sender] != NULL)
	{
	  ports[sender]->close();
	  delete ports[sender];
	  ports[sender] = NULL;
	}
      break;
    }
  
  return 0;
}

int BV_AudioApp::playADPCM(int mbr, char *data, int len)
{
  char buffer[2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE];
  struct adpcm_state adpcmstate;

  memcpy(&adpcmstate.valprev, data, 2);
  adpcmstate.index = data[2];
  data += 3;
  adpcm_decoder(data, (short *)buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE, 
                &adpcmstate);

  //session_->lock();
  //((MainForm *)main_form_)->setSpeakerLevel
  //  (get_level(buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE));
    //session_->unlock();

  ports[mbr]->play(buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

  return 0;
}

int BV_AudioApp::playGSM(int mbr, char *data, int len)
{
  char buffer[AUDIO_CHUNKS_NO][320];
  int i;
  
  for(i=0; i<AUDIO_CHUNKS_NO; i++)
    {
      gsm_decode(gsm_handle_, (gsm_byte *)((gsm_frame *)data)[i], 
		 (gsm_signal *)buffer[i]);
    }
  
  //session_->lock();
  //((MainForm *)main_form_)->setSpeakerLevel
  //(get_level((char *)buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE));
    //session_->unlock();

  ports[mbr]->play((char *)buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

  return 0;
}

int BV_AudioApp::send_audio()
{
  int size = 0;
  static int seq = 0;

  //session_->lock();
  BV_Member *mbr = session_->get_member(session_->get_local_id());
  //session_->unlock();
  if((state_ == START) && (mbr != NULL))
    {
      switch(compression_)
	{
	case PCM:
	  size = 1600;
	  sendPCM();
	  break;

	case ADPCM:
	  size = 400;
	  sendADPCM();
	  break;

	case GSM:
	  size = 33*5;
	  sendGSM();
	  break;
	}

      RTPheader header;
      header.seq = seq;
      seq = (seq + 1) % 65555;
      //session_->lock();
      mbr->update_audio_stat(&header, size);
      //session_->unlock();
      //session_->update_audio_stat(mbr);
    }
  else
    {
      sleep(1);
      return -1;
    }

  return 0;
}

int BV_AudioApp::sendPCM()
{
  char buffer[2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE];
  int level;

  mic_port_->capture(buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

  level = get_level(buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
  //((MainForm *)main_form_)->setMicLevel(level);
  level = level * 200 / 65536;

  if(level >= mic_cutoff)
    {
      send_data(PCM, buffer, 2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
    }

  return 0;
}

int BV_AudioApp::sendADPCM()
{
  char buffer[2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE];
  char adpcm[AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE/2 + 3];
  char *tmp;
  int size;
  int level;
  
  mic_port_->capture(buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

  level = get_level(buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
  //((MainForm *)main_form_)->setMicLevel(level);
  level = level * 200 / 65536;

  if(level >= mic_cutoff)
    {
      memcpy(adpcm, &adpcmsendstate.valprev, 2);
      adpcm[2] = adpcmsendstate.index;
  
      tmp = adpcm + 3;
      adpcm_coder((short *)buffer, tmp, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE, &adpcmsendstate);
      
      session_->lock();
      send_data(ADPCM, adpcm, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE/2 + 3);
      session_->unlock();
    }

  return 0;
}

int BV_AudioApp::sendGSM()
{
  char buffer[AUDIO_CHUNKS_NO][320];
  gsm_frame frames[AUDIO_CHUNKS_NO];
  int size;
  int i;
  int level;

  mic_port_->capture((char *)buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

  level = get_level((char *)buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
  //session_->lock();
  //((MainForm *)main_form_)->setMicLevel(level);
  //session_->unlock();
  level = level * 200 / 65536;

  if(level >= mic_cutoff)
    {
      for(i=0; i<AUDIO_CHUNKS_NO; i++)
	{
	  gsm_encode(gsm_handle_, (gsm_signal *)buffer[i], (gsm_byte *)&frames[i]);
	}
      
      send_data(GSM, (char *)frames, AUDIO_CHUNKS_NO*sizeof(gsm_frame));
    }

  return 0;
}

short BV_AudioApp::get_level(char *buffer, int no)
{
  int i;
  short max = 0;
  
  for(i = 0; i < no; i++)
    {
      if(((short *)buffer)[i] > max)
	max = ((short *)buffer)[i];
    }
  return max;
}

void BV_AudioApp::update_options(Options *options)
{
  switch(options->connection)
    {
    case _modem28:
      compression_ = GSM;
      break;

    case _isdn:
      compression_ = ADPCM;
      break;
      
    case _isdn2:
      compression_ = ADPCM;
      break;

    case _lan:
      compression_ = PCM;
      break;

    default:
      compression_ = GSM;
      break;
    }

  BV_AudioPlay::set_volume(options->volume);
  BV_AudioCapture::set_gain(options->mic_volume);

  mic_cutoff = options->mic_cutoff;
}

void BV_AudioApp::play()
{
  state_ = START;
}

void BV_AudioApp::stop()
{
  state_ = STOP;
}
